"""
File/modulo contenente la classe "Deployer"
"""
import os
import stat
import shutil
import subprocess
import re
from tempfile import NamedTemporaryFile


class ActionNotFoundError(Exception):
    """ Eccezione sollevata quando viene richiesta un'azione non esistente """
    pass


class Deployer:
    """
    Classe che si occupa del deployment sul server
    """
    __config = {}
    __placeholder_format = '##{}##'
    __default_actions_shell = '/bin/bash'       # shell da utilizzare di default nel caso non venga comunicata nella config
    __default_actions_timeout = 180             # timeout in secondi da utilizzare nel caso che non venga specificato niente nella config
    __permission_mask = 0o777

    def __init__(self, configuration):
        """
        Inizializzazione dello stato interno
        :param configuration:          configurazione del deployer
        """
        self.__config = configuration

        for s in {'action', 'folder'}:
            if s not in self.__config:
                self.__config[s] = {}

    def __run(self, bash_string):
        """
        Funzione che esegue i comandi specificati in bash_string e ritorna il loro output
        :param bash_string:         stringa contente i comandi (attenzione ai new_line)
        :return:                    dizionario contenente l'exit_code e l'exit_message
        """

        # ottengo la shell da utilizzare per le azioni
        actions_shell = self.__config.get('DEFAULT', {}).get('actions-shell', self.__default_actions_shell)
        actions_timeout = self.__config.get('DEFAULT', {}).get('actions-timeout', self.__default_actions_timeout)

        # si deve crare un file bash contenente tutti i comandi in modo che vengano eseguiti nello stesso enviroment
        # creo un file temporaneo in /tmp
        bash_script = NamedTemporaryFile(mode='w', delete=False)
        # scrivo i comandi sul file
        bash_script.write(bash_string)
        bash_script.close()

        # modifico i permessi del file in modo che sia eseguibile
        os.chmod(bash_script.name, stat.S_IEXEC | stat.S_IREAD)
        # eseguo lo script chiamandolo utilizzando la shell specificata e ridirigo:
        # il suo stderr verso lo stdout, poichè voglio avere entrambi gli stream assieme, nell'ordine dei messaggi giusto
        # il suo stdin verso /dev/null, poichè se qualcuno nello script cerca di chiamare un programma che legge dallo standard input qui hanga forever
        proc_handle = subprocess.Popen([actions_shell, bash_script.name], stdin=subprocess.DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        # leggo lo standard output del processo, tenendo conto del timeout
        try:
            stdout, _ = proc_handle.communicate(timeout=actions_timeout)
        except subprocess.TimeoutExpired:
            # se il timeout del processo scade, lo killo e leggo il suo stdout
            proc_handle.kill()
            # qui si può piantare solo nel caso che il processo stia eseguendo una system call piantata a causa di un fattore esterno.
            # un esempio è la lettura da un disco rimovibile rotto, o un modulo kernel buggato
            stdout, _ = proc_handle.communicate()

        return_code = proc_handle.returncode

        # elimino il file
        os.unlink(bash_script.name)

        return {'exit_code': return_code, 'exit_message': stdout.decode()}

    @staticmethod
    def __get_common_prefix_len(path, directory, reflective):
        """
        Funzione che controlla se un path ricade all'interno di una directory
        Se reflective è vero, Non è riflessiva, quindi una cartella non è all'interno di se stessa
        :param path:                path che deve ricadere
        :param directory:           directory nella quale deve ricadere
        :param reflective           indica se la funzione deve essere riflessiva, ovvero se una cartella viene considerata all'interno di se stessa
        :return:                    la lunghezza del prefisso comune tra il path e la cartella
        """
        # realpath elimina i link simbolici all'interno del path
        directory = os.path.realpath(directory).rstrip('/') + '/'
        path = os.path.realpath(path) + ('/' if reflective else '')

        # controllo se il prefisso comune ai due path
        if os.path.commonprefix([directory, path]) == directory:
            return len(directory)
        else:
            return 0

    def __resolve_path_aliases(self, path):
        """
        Funzione che dato un path risolve eventuali alias al suo interno e restituisce il path assoluto
        Se il path passato non dovesse contenere aliases allora viene restituito invariato
        :param path:                path nel quale risolvere gli aliases
        :return:                    path con gli aliases risolti
        """
        # aliases => dizionario con chiave = nome placeholder, valore = valore da sostituire
        for alias in self.__config.get('aliases', {}):
            # per ogni alias definito nel dizionario degli aliases verifico se è presente nel path e in caso lo sostituisco
            # il formato del placeholder è definito come variabile privata della classe
            placeholder = self.__placeholder_format.format(alias)
            path = path.replace(placeholder, self.__config['aliases'][alias])

        return os.path.abspath(path)

    def __get_root_folder(self, path, reflective=False):
        """
        Funzione che verifica se un path ricade tra quelli che sono modificabili e restituisce la directory che matcha
        in modo più specifico
        :param path:            path da verificare
        :param reflective       indica se la funzione deve considerare le cartelle come riflessive, ovvero se una cartella viene considerata all'interno di se stessa
        :return:                il path all'interno del quale ricade
        """

        max_common_prefix_len = 0
        best_matching_folder = None

        # scorro tutte le cartelle che sono modificabili da questo deployer
        for directory in self.__config['folder']:
            common_prefix_len = self.__get_common_prefix_len(path=path, directory=directory, reflective=reflective)
            if common_prefix_len > max_common_prefix_len:
                max_common_prefix_len = common_prefix_len
                best_matching_folder = directory

        return best_matching_folder

    def __validate_permissions(self, permissions):
        """
        Funzione di controllo dei permessi passati
        Viene controllato che siano del tipo giusto e viene applicata la maschera
        :param permissions:  valore passato come argomento del parametro indicante i permessi
        :return:             permessi
        """

        if not isinstance(permissions, int):
            raise RuntimeError('Permissions must be specified as an octal number (eg. 0o777, 0o0777, 0777, 777, etc.')

        # maschero i permessi per eliminare configurazioni non ammissibili
        perms = permissions & self.__permission_mask

        # controllo se i permessi sono stati modificati dalla maschera
        if perms != permissions:
            raise RuntimeError('Permissions have been modified by the mask. Please change your permissions.')

        return perms

    def __get_perms_of_root_folder(self, root_folder):
        """
        Funzione che ritorna i permessi definiti per una cartella root presente nella config di questo deployer
        :param root_folder:         cartella root della quale si vogliono ottenere i permessi
        :return:                    permessi cartelle, permessi files, uid, gid
        """
        dir_perms = self.__config['folder'][root_folder]['dir-perms']
        file_perms = self.__config['folder'][root_folder]['file-perms']

        # ottengo l'uid dell'owner e il gid del group
        uid = self.__config['folder'][root_folder]['owner']
        gid = self.__config['folder'][root_folder]['group']

        return dir_perms, file_perms, uid, gid

    def execute_action(self, action_name):
        """
        Esecuzione dell'azione specificata
        :param action_name:     nome dell'azione che si vuole eseguire
        :throw ActionNotFoundError:  nel caso non venga trovata l'azione richiesta
        :throw RuntimeError:    nel caso ci siano problemi con l'esecuzione
        :return:                dizionario contenente l'exit_code e l'exit_message
        """
        # controllo se l'azione esiste per questo deployer
        if action_name not in self.__config['action']:
            raise ActionNotFoundError("Cannot find action \"{}\"".format(action_name))

        # ottengo la lista di comandi da eseguire per questa azione
        commands = self.__config['action'][action_name]

        # eseguo i comandi
        try:
            results = self.__run(commands)
        except RuntimeError as e:
            raise RuntimeError('Something went wrong running action "{}". {}.'.format(action_name, e))

        # ritorno la lista degli output dei comandi
        return results

    def delete_filesystem_node(self, to_delete_path):
        """
        Funzione che dato un path elimina il nodo del filesystem corrispondente
        :param to_delete_path:      path del nodo da eliminare
        :throw PermissionError:     nel caso il nodo non rientri in un path modificabile dal deployer
        :throw RuntimeError:        nel ci siano dei problemi con l'eliminazione del nodo
        :throw FileNotFoundError:   nel caso il nodo non venga trovato
        """
        # ottengo il path pulito (rimuovo gli alias e risolvo i link)
        to_delete_path = self.__resolve_path_aliases(path=to_delete_path)

        # controllo se il path è accessibile
        if self.__get_root_folder(path=to_delete_path) is None:
            # il path non è modificabile, quindi restituisco errore
            raise PermissionError('Target file is located in a not reachable path!')

        # se il path ricade tra quelli modificabili, eseguo l'azione
        try:
            if os.path.isfile(to_delete_path):
                # se il path punta a un file allora rimuovo il file
                os.remove(to_delete_path)
            elif os.path.isdir(to_delete_path):
                # se il path punta a una directory allora rimuovo ricorsivamente tutto il suo contenuto
                shutil.rmtree(to_delete_path)
            else:
                # la tipologia non è tra quelle eliminabili
                raise FileNotFoundError('Path given as argument "{}" is not a file and not a directory!'.format(to_delete_path))
        except OSError as e:
            # errore durante l'eliminazione
            raise RuntimeError('An error occurred deleting file or directory "{}". OSError: {}.'.format(to_delete_path, e))

    def delete_by_regex(self, base_path, regex, inverted, node_type, include_subdirs):
        """
        Funzione che dato un path elimina il nodo del filesystem corrispondente
        :param base_path:          path della cartella da cui cominciare ad eliminare
        :param regex:              regex con cui verificare se eliminare un elemento
        :param inverted:           booleano che indica se invertire il matche effettuato dalla regex
        :param node_type:          indica se considerare solo file, solo cartelle o entrambi nell'eliminazione
        :param include_subdirs:    indica se considerare solo la directory corrente o anche le sottodirectory per le eliminazioni

        :throw PermissionError:     nel caso il base path non rientri in un path modificabile dal deployer
        :throw RuntimeError:        nel ci siano dei problemi con l'eliminazione del nodo
        :throw FileNotFoundError:   nel caso il base path non venga trovato
        """

        # ottengo il path pulito (rimuovo gli alias e risolvo i link)
        base_path = self.__resolve_path_aliases(path=base_path)

        # controllo se il path è accessibile
        if self.__get_root_folder(path=base_path, reflective=True) is None:
            # il path non è modificabile, quindi restituisco errore
            raise PermissionError('Target base path is located in a not reachable path!')

        # se il path ricade tra quelli modificabili, eseguo l'azione

        try:
            r = re.compile(regex)
        except re.error as e:
            # errore durante la preparazione della regex
            raise RuntimeError('An error occurred compiling the regex "{}": {}.'.format(regex, e))

        node_type = node_type.lower().strip()

        include_files = (node_type in {'file', 'both'})
        include_folders = (node_type in {'folder', 'both'})

        errors, too_many_errors = self.__delete_in_folder(base_path, r, inverted, include_files, include_folders, include_subdirs)

        return errors, too_many_errors


    def __delete_in_folder(self, base_path, regex, inverted, include_files, include_folders, include_subdirs):

        if self.__get_root_folder(path=os.path.realpath(base_path), reflective=True) is None:
            # in qualche modo ci troviamo a cercare di eliminare in una directory che è fuori dai path permessi (a causa di un symlink per esempio)
            # ritorniamo immediatamente poichè non possiamo proseguire
            return [], False

        errors = []
        too_many_errors = False

        # gestisco le entry in questa cartella
        for entry in os.listdir(base_path):
            # scorro tutte le entry presenti all'interno della cartella

            # ottengo il path assoluto dell'entry
            entry_abs_path = os.path.join(base_path, entry)

            # salto le entry che non mi interessano
            if os.path.islink(entry_abs_path) or os.path.ismount(entry_abs_path):
                continue

            try:
                if os.path.isdir(entry_abs_path):
                    # LA ENTRY È UNA DIRECTORY

                    # se le directory devono essere considerate
                    if include_folders:
                        # verifico se matcha la regex
                        m = regex.fullmatch(entry) is not None

                        if (m and not inverted) or (not m and inverted):
                            # elimino la directory
                            shutil.rmtree(entry_abs_path)


                elif os.path.isfile(entry_abs_path):
                    # LA ENTRY È UN FILE

                    # se i file devono essere considerati
                    if include_files:
                        # verifico se matcha la regex
                        m = regex.fullmatch(entry) is not None

                        if (m and not inverted) or (not m and inverted):
                            # elimino il file
                            os.remove(entry_abs_path)

            except OSError as e:
                # errore durante l'eliminazione,
                if not too_many_errors:
                    errors.append([entry_abs_path, "{}".format(e)])
                    if len(errors) > 100:
                        too_many_errors = True


        # se vanno considerate anche le sottodirectory
        if include_subdirs:
            for d in os.listdir(base_path):
                if os.path.isdir(os.path.join(base_path, d)):
                    tmp_errors, tmp_too_many = self.__delete_in_folder(os.path.join(base_path, d), regex, inverted, include_files, include_folders, include_subdirs)

                    # se non abbiamo già raggiunto i 100 errori
                    if len(errors) < 100:
                        # aggiungo gli errori fino ad un massimo di 100
                        errors += tmp_errors[0:100-len(errors)]

                    # too_many_errors è vero se lo era prima oppure se questa iterazione lo riporta come vero
                    too_many_errors = too_many_errors or tmp_too_many

        return errors, too_many_errors

    def create_folder(self, folderpath, overwrite=False, permissions=None):
        """
        Funzione che dato il path di una cartella la crea
        :param folderpath:              path della cartella da creare
        :param overwrite:               flag che indica se la cartella deve essere sovrascritta
        :param permissions:             permessi da assegnare alla cartella
        :throw PermissionError:         nel caso la cartella non rientri in un path modificabile dal deployer
        :throw FileExistsError:         nel caso la cartella esista già (per poter gestire/skippare il problema dal chiamante)
        :throw RuntimeError:            nel caso ci siano problemi con la creazione della cartella
        """
        # ottengo il path pulito (rimuovo gli alias e risolvo i link)
        folderpath = self.__resolve_path_aliases(path=folderpath)

        # ottengo la cartella parent definita nella configurazione per sapere che permessi applicare
        root_folder = self.__get_root_folder(path=folderpath)

        if root_folder is None:
            # il path non è modificabile, quindi restituisco errore
            raise PermissionError('Target folder is located in a not reachable path!')

        # controllo se la cartella esiste già
        if os.path.isdir(folderpath) and not overwrite:
            raise FileExistsError('Folder "{}" already exists!'.format(folderpath))
        elif os.path.isdir(folderpath) and overwrite:
            # la cartella esiste e il flag overwrite è attivo, quindi elimino la vecchia cartella
            try:
                shutil.rmtree(folderpath)
            except Exception as e:
                # c'è stato un errore nell'eliminazione, lo segnalo al chiamante
                raise RuntimeError(e)
        elif not os.path.isdir(os.path.dirname(folderpath)):
            # la cartella nella quale si vuole creare la cartella non esiste
            raise RuntimeError('Parent folder "{}" does not exist.'.format(os.path.dirname(folderpath)))

        # ottengo i permessi
        dir_perms, _, uid, gid = self.__get_perms_of_root_folder(root_folder=root_folder)

        perms = dir_perms if permissions is None else self.__validate_permissions(permissions=permissions)

        # effettuo le modifiche all'umask per poter impostare in modo corretto i permessi
        umask = 0o777 ^ perms
        umask_original = os.umask(umask)

        try:
            # creo la cartella specificandone i vari permessi
            os.mkdir(path=folderpath, mode=perms)
            os.chown(path=folderpath, uid=uid, gid=gid)
        except OSError as e:
            raise RuntimeError('An error occurred creating folder "{}". OSError: {}.'.format(folderpath, e))
        finally:
            # sia che la creazione abbia avuto successo o meno, ripristino l'umask al suo stato iniziale
            os.umask(umask_original)

    def get_file_uploader(self, filepath, overwrite=False, permissions=None):
        """
        Funzione per la creazione di un uploader di files
        :param filepath:            path del file che si vuole creare
        :param overwrite:           indica se nel caso il file esista già debba venire sovrascritto
        :param permissions:         permissions da assegnare al file
        :throw PermissionError:     se la cartella in cui si vuole creare il file non è accessibile
        :throw FileExistsError:     nel caso il file esista già e overwrite è impostato a False
        :throw RuntimeError:        nel caso ci siano dei problemi con la creazione del file
        :return: handle, str        handle del file sul quale poter chiamare le operazioni di scrittura, path effettivo del file aperto
        """
        # ottengo il path pulito (rimuovo gli alias e risolvo i link)
        new_file = self.__resolve_path_aliases(path=filepath)

        # ottengo la cartella parent definita nella configurazione per sapere che permessi applicare
        root_folder = self.__get_root_folder(path=new_file)

        if root_folder is None:
            # il path non è modificabile, quindi restituisco errore
            raise PermissionError('File to be created is located in a not reachable path!')

        if not os.path.isdir(os.path.dirname(new_file)):
            # la cartella nella quale si vuole creare il file non esiste
            raise RuntimeError('Parent folder "{}" does not exist.'.format(os.path.dirname(new_file)))

        if os.path.isfile(new_file) and overwrite:
            # il file esiste già ed overwrite è attivo, allora cancello il vecchio file
            try:
                os.remove(new_file)
            except OSError as e:
                # segnalo il problema con l'eliminazione del file
                raise RuntimeError("An exception occurred removing old file: (OSError) {}".format(e))
        elif os.path.isfile(new_file) and not overwrite:
            # segnalo che il file esisteva
            raise FileExistsError("File already exists and overwrite is set to False.")

        # ottengo i permessi della cartella nella quale verrà posizionato il file
        _, file_perms, uid, gid = self.__get_perms_of_root_folder(root_folder=root_folder)

        perms = file_perms if permissions is None else self.__validate_permissions(permissions=permissions)

        # importazione di alcuni flag utili
        # il file viene creato se non esiste già (O_CREAT) se esiste già viene sollevata un'eccezione (O_EXCL)
        flags = os.O_CREAT | os.O_WRONLY | os.O_TRUNC | os.O_EXCL

        # modifico l'umask a seconda dei permessi che ho intenzione di impostare al nuovo file
        umask = 0o777 ^ perms
        umask_original = os.umask(umask)

        try:
            # creo il file e imposto immediatamente l'owner e il group
            file_descriptor = os.open(new_file, flags, perms)
            os.fchown(fd=file_descriptor, uid=uid, gid=gid)
        except Exception as e:
            raise RuntimeError('An error occurred opening file: {}'.format(e))
        finally:
            # torno a ripristinare l'umask in ogni caso
            os.umask(umask_original)

        return os.fdopen(file_descriptor, 'wb'), new_file


import os
import grp
import pwd

"""
Implementazione di un parser per il formato del file di configurazione simil-INI usato dall'applicazione.
"""


class ConfigFile(object):
    """
    Classe per il parsing delle configurazioni di deployment presenti sul server
    """
    # nome del file di configurazione rappresentato dalla classe
    __filename = None
    # contenuto della configurazione
    __config = None
    # formato del placeholder
    __placeholder_format = '##{}##'

    def __init__(self, filename):
        """
        Inizializzazione della configurazione e del suo stato interno
        :throw FileNotFoundError:       nel caso il file non venga trovato
        :throw RuntimeError:            nel caso ci siano problemi con gli aliases
        :throw ValueError:              nel caso la configurazione segua la struttura definita
        """

        if not os.path.isfile(filename):
            # il file non esiste, sollevo un'eccezione
            raise FileNotFoundError('Il file della configurazione non è stato trovato')

        # imposto il nome del file e lo leggo
        self.__filename = filename
        data = open(filename, 'r').read()

        # parso il file di configurazione, allo stato attuale la configurazione non è detto che sia consistente
        unvalidated_config = self.__parse(data)                         # può sollevare un ValueError

        # risolvo gli alias della configurazione
        self.__resolve_config_aliases(unvalidated_config)               # può sollevare un RuntimeError

        # verifico e fixo la configurazione e imposto effettivamente lo stato interno
        self.__config = self.__validate_and_fix(unvalidated_config)     # può sollevare un ValueError

    @staticmethod
    def __parse(data):
        """
        Funzione per parsare la configurazione e ottenere un dizionario
        :param data:            Dati grezzi letti dal file di config
        :throw ValueError:      se i separatori o le sezioni non sono state riconosciute
        :return: Dizionario che rappresenta il contenuto del file di config, parsato
        """
        # dizionario che conterrà la configurazione
        conf = {}

        # variabili per tenere in memoria la sezione corrente
        current_section = None
        current_subsection = None

        # separa il file in righe
        for l in data.split('\n'):
            # elimina gli spazi iniziali e finali
            l = l.strip('\r\t ')

            # ignora le righe inutili, ovvero i commenti e le righe bianche
            if len(l) == 0 or l.startswith('#'):
                continue

            # SEZIONI E SOTTOSEZIONI
            if l.startswith('['):
                # gestisco gli inizi di sezione
                if l.lower().startswith(('[aliases]', '[default]')):
                    current_section, current_subsection = l.lower().strip('[]'), None
                elif l.lower().startswith(('[folder:', '[action:')):
                    l = l.strip('[]').split(':')
                    current_section, current_subsection = l[0].lower().strip('[]'), ':'.join(l[1:]).strip()
                else:
                    raise ValueError("Il separatore di sezione '{}' non è valido.".format(l))

                # creo le sezioni e sottosezioni se non esistono
                if current_section not in conf:
                    conf[current_section] = {}

                if current_subsection is not None and current_subsection not in conf[current_section]:
                    if current_section in ['action']:
                        # se è una sottosezione di action allora conterrà una stringa
                        conf[current_section][current_subsection] = ''
                    else:
                        # altrimenti conterrà un dizionario chiave - valore
                        conf[current_section][current_subsection] = {}
                # se la riga conteneva una sezione o sottosezione ho finito l'elaborazione e vado alla prossima
                continue


            # gestisco le righe che sono contenute nelle sezioni, in base alla sezione che le contiene
            # se siamo arrivati qui, significa che stiamo gestendo una riga che non è un separatore di sezione

            # sezioni che usano la struttura chiave = valore
            if current_section in ['default', 'aliases', 'folder']:
                l = l.split('=')
                key, value = l[0].strip(), '='.join(l[1:]).strip()

            # sezioni che usano la struttura valore e basta
            elif current_section in ['action']:
                key, value = l, None

            else:
                # la sezione non è stata riconosciuta
                raise ValueError("La sezione '{}' è sconosciuta.".format(current_section))


            # aggiungo la chiave e valore decodificati alla conf finale
            if current_section in ['action']:
                # aggiungo il comando alla lista dei comandi da eseguire per questa azione
                # new line necessario per la separazione dei comandi
                conf[current_section][current_subsection] += '\n{}'.format(key)

            elif current_section in ['default']:
                # gestione delle chiavi multivalore, i valori vengono aggiunti alla lista
                if key not in conf[current_section]:
                    # prima volta che incontro la chiave, inizializzo la lista
                    conf[current_section][key] = []
                conf[current_section][key].append(value)

            elif current_section in ['aliases']:
                # gli aliases vengono risolti con un dizionario
                if key not in conf[current_section]:
                    conf[current_section][key] = {}

                conf[current_section][key] = value

            elif current_section in ['folder']:
                if key not in conf[current_section][current_subsection]:
                    conf[current_section][current_subsection][key] = {}

                conf[current_section][current_subsection][key] = value

        return conf

    def __validate_and_fix(self, data):
        """
        Validazione della configurazione e piccoli fix automatici
        :param data:            Dizionario parsato da parte e che deve essere validato
        :throw ValueError:      Sezioni mancanti, non ammesse, permessi o utenti non validi
        :return:                Dizionario come definito per le configurazioni
        """

        # verifico che le sezioni necessarie esistano
        needed_sections = {'default'}
        if not all(sec in data for sec in needed_sections):
            raise ValueError("Le sezioni obbligatorie ({}) non sono tutte presenti. Nel file di configurazione '{}'".format(needed_sections, self.__filename))

        # verifico che le sezioni esistenti siano dentro quelle ammesse
        allowed_sections = {'default', 'aliases', 'folder', 'action', 'owner', 'group', 'dir-perms', 'file-perms'}
        if not all(sec in allowed_sections for sec in data):
            raise ValueError("Alcune delle sezioni specificate non sono ammesse, sezioni valide: {}. Nel file di configurazione '{}'".format(allowed_sections, self.__filename))

        # verifico che se sottosezioni siano valide

        # verifico che le chiavi necessarie esistano in DEFAULT
        needed_keys = {'key'}           # il campo dell'apikey non può mancare
        if not all(sec in data['default'] for sec in needed_keys):
            raise ValueError("Le chiavi obbligatorie ({}) per la sezione '[{}]' non sono tutte presenti. Nel file di configurazione '{}'".format(needed_keys, 'default', self.__filename))

        # verifico che le chiavi esistenti siano dentro quelle ammesse in DEFAULT
        allowed_keys = {'key', 'allow-ip', 'owner', 'group', 'dir-perms', 'file-perms', 'actions-shell', 'actions-timeout'}
        if not all(sec in allowed_keys for sec in data['default']):
            raise ValueError("Alcune delle chiavi specificate non sono ammesse, chiavi valide: {} nella sezione '[{}]'. Nel file di configurazione '{}'".format(allowed_keys, 'default', self.__filename))

        # trasformo le direttiva da liste al tipo corretto

        for k in data['default'].keys():
            # trasformazione delle liste in sets in modo da non aver duplicati
            if k in {'allow-ip'}:
                data['default'][k] = set(data['default'][k])

            elif k in {'dir-perms', 'file-perms'}:
                # verifico che i permessi specificati siano validi
                if self.__are_permissions_valid(data['default'][k][0]):
                    data['default'][k] = data['default'][k][0]
                else:
                    raise ValueError("I permessi directory '{}' non sono validi. Nella sezione '[{}]'. Nel file di configurazione '{}'.".format(data['default'][k][0], 'default', self.__filename))

            elif k in {'owner'}:
                # verifico che owner e group sia esistente

                if self.__to_user_id(data['default'][k][0]) is not None:
                    data['default'][k] = data['default'][k][0]
                else:
                    raise ValueError("L'utente '{}' non è valido. Nella sezione '[{}]'. Nel file di configurazione '{}'.".format(data['default'][k][0], 'default', self.__filename))

            elif k in {'group'}:
                # verifico che group sia esistente

                if self.__to_group_id(data['default'][k][0]) is not None:
                    data['default'][k] = data['default'][k][0]
                else:
                    raise ValueError("Il utente '{}' non è valido. Nella sezione '[{}]'. Nel file di configurazione '{}'.".format(data['default'][k][0], 'default', self.__filename))

            else:
                # per ogni altro campo, nel caso ci fossero più valori allora tengo il primo
                data['default'][k] = data['default'][k][0]


        # verifico che le chiavi necessarie esistano in FOLDER
        # tolgo dalle chiavi richieste quelle già specificate nella sezione default
        needed_keys = {'group', 'owner', 'dir-perms', 'file-perms'} - set(data['default'].keys())
        for f in data['folder']:
            if not os.path.isabs(f):
                raise ValueError("La sezione '[{}:{}]' descrive un percorso non assoluto. I percorsi devono essere assoluti. Nel file di configurazione '{}'.".format('folder', f, self.__filename))

            if not os.path.isdir(f):
                raise ValueError("La sezione '[{}:{}]' non descrive una directory oppure è inesistente. I percorsi devono essere assoluti. Nel file di configurazione '{}'.".format('folder', f, self.__filename))

            if not all(sec in data['folder'][f] for sec in needed_keys):
                raise ValueError("Le chiavi obbligatorie ({}) per la sezione '[{}:{}]' non sono tutte presenti. Aggiungerle nella sezione specificata o nella sezione di default. Nel file di configurazione '{}'".format(needed_keys, 'folder', f, self.__filename))

        # verifico che le chiavi esistenti siano dentro quelle ammesse
        allowed_keys = {'group', 'owner', 'dir-perms', 'file-perms'}
        for f in data['folder']:
            if not all(sec in allowed_keys for sec in data['folder'][f]):
                raise ValueError("Alcune delle chiavi specificate non sono ammesse, chiavi valide: {} nella sezione '[{}:{}]'. Nel file di configurazione '{}'".format(allowed_keys, 'folder', f, self.__filename))

        # aggiungo alle singole folder i permessi ereditati dal master
        needed_keys = {'group', 'owner', 'dir-perms', 'file-perms'}
        for f in data['folder']:
            for nk in needed_keys:
                if nk not in data['folder'][f]:
                    data['folder'][f][nk] = data['default'][nk]

        # verifico che le folder abbiano permessi validi
        for f in data['folder']:

            if self.__are_permissions_valid(data['folder'][f]['dir-perms']):
                data['folder'][f]['dir-perms'] = int(data['folder'][f]['dir-perms'], 8)
            else:
                raise ValueError("I permessi directory '{}' non sono validi. Nella sezione '[{}:{}]'. Nel file di configurazione '{}'.".format(data['folder'][f]['dir-perms'], 'folder', f, self.__filename))

            if self.__are_permissions_valid(data['folder'][f]['file-perms']):
                data['folder'][f]['file-perms'] = int(data['folder'][f]['file-perms'], 8)
            else:
                raise ValueError("I permessi directory '{}' non sono validi. Nella sezione '[{}:{}]'. Nel file di configurazione '{}'.".format(data['folder'][f]['file-perms'], 'folder', f, self.__filename))

            gid = self.__to_group_id(data['folder'][f]['group'])

            if gid is not None:
                data['folder'][f]['group'] = gid
            else:
                raise ValueError("Il gruppo '{}' non è valido. Nella sezione '[{}:{}]'. Nel file di configurazione '{}'.".format(data['folder'][f]['group'], 'folder', f, self.__filename))

            uid = self.__to_user_id(data['folder'][f]['owner'])

            if uid is not None:
                data['folder'][f]['owner'] = uid
            else:
                raise ValueError("Il gruppo '{}' non è valido. Nella sezione '[{}:{}]'. Nel file di configurazione '{}'.".format(data['folder'][f]['owner'], 'folder', f, self.__filename))


        # elimino i default dalla configurazione, ora che hanno terminato il loro scopo
        data['default'].pop('owner', None)
        data['default'].pop('group', None)
        data['default'].pop('dir-perms', None)
        data['default'].pop('file-perms', None)


        return data

    def __resolve_config_aliases(self, config_dict):
        """
        Procedura che risolve gli alias all'interno della configurazione
        Da chiamare in seguito alla lettura e validazione della configurazione

        Gli alias vengono risolti all'interno delle sezioni: folder e actions

        :throw RuntimeError:            nel caso in seguito alla risoluzione degli alias due cartelle siano duplicate
        """

        # inizio risolvendo quelli all'interno delle sezioni folders
        for old_folder, old_folder_value in config_dict.get('folder', {}).copy().items():
            # scorro le chiavi, i nomi delle cartelle sono nelle chiavi e non nei valori
            # ottengo il nome della cartella con gli alias risolti
            new_folder = os.path.abspath(self.__resolve_aliases(old_folder, config_dict))

            # devo rimuovere la chiave (salvando il suo valore) in modo che se non sono stati risolti alias, non entri in conflitto con se stessa
            config_dict['folder'].pop(old_folder)

            if new_folder in config_dict['folder']:
                raise RuntimeError('Durante la risoluzione degli alias sono state ottenute due cartelle duplicate. '
                                   'La cartella "{}" viene risolta in "{}" che è già presente'.format(old_folder, new_folder))
            else:
                # vado a sostituire la chiave all'interno del dizionario
                config_dict['folder'][new_folder] = old_folder_value

        # ora risolvo gli alias all'interno delle azioni
        for action_name, old_action in config_dict.get('action', {}).copy().items():
            # scorro le azioni, i placeholder devono essere sostituiti all'interno delle stringhe delle azioni
            action = "".join(old_action)            # uso un join in modo da poter gestire allo stesso modo stringhe e liste di stringhe

            # risolvo gli alias nell'azione e la sostituisco
            new_action = self.__resolve_aliases(action, config_dict)
            config_dict['action'][action_name] = new_action

    def __resolve_aliases(self, path, config_dict):
        """
        Funzione che dato un path risolve eventuali alias al suo interno e restituisce il path assoluto
        Se il path passato non dovesse contenere aliases allora viene restituito invariato

        :param path:                path nel quale risolvere gli aliases
        :throw RuntimeError:        nel caso non esiste un alias con il quale risolvere un placeholder

        :return:                    path con gli aliases risolti
        """
        # aliases => dizionario con chiave = nome placeholder, valore = valore da sostituire
        for alias in config_dict.get('aliases', []):
            # mettere uno slash in coda
            # per ogni alias definito nel dizionario degli aliases verifico se è presente nel path e in caso lo sostituisco
            # il formato del placeholder è definito come variabile privata della classe
            placeholder = self.__placeholder_format.format(alias)
            path = path.replace(placeholder, config_dict['aliases'][alias] + '/')

        return path

    def to_dict(self):
        return self.__config

    @staticmethod
    def __are_permissions_valid(permissions):
        """
        Verifica che dei permessi in formato ottale siano validi
        :param permissions: Stringa che contiene i permessi in ottale da validare
        """
        try:
            # converto da ottale i permessi
            int(permissions, 8)
        except ValueError:
            return False

        return True

    @staticmethod
    def __to_user_id(user):
        """
        Converte un nome di un gruppo nel relativo id
        :param user: Nome dell'utente
        :return: Id linux del gruppo
        """
        try:
            # ottengo l'id del gruppo
            uid = pwd.getpwnam(user).pw_uid
        except KeyError:
            return None

        return uid

    @staticmethod
    def __to_group_id(group):
        """
        Converte un nome di un gruppo nel relativo id
        :param group: Nome del gruppo
        :return: Id linux del gruppo
        """
        try:
            # ottengo l'id del gruppo
            gid = grp.getgrnam(group).gr_gid
        except KeyError:
            return None

        return gid


SERVER_CONF_EXAMPLE = """
[global]

# connection info
server.socket_host: "0.0.0.0"
server.socket_port: 47000
server.thread_pool: 8
server.socket_timeout: 60
response.timeout: 3600
server.max_request_body_size: 0

# ad hoc max request size
lisp.request.max_size_mb: 2

# max size for file uploads
lisp.file_upload.max_size_mb: 200

# chunk size for file upload
lisp.file_upload.chunk_size_mb: 5

# ssl (to enable ssl de-comment lines below and replace placeholders)
#server.ssl_certificate: '< PEM CERT ABS. PATH>'
#server.ssl_private_key: '< PEM KEY ABS. PATH>'

# disabilitazione dell'auto-reload
engine.autoreload.on: False

# set log files
lisp.logging.cherrypy_access.file: '< ACCESS LOG FILE ABS. PATH >'
lisp.logging.cherrypy_error.file: '< ERROR LOG FILE ABS. PATH >'
lisp.logging.cherrypy_app.file: '< APP LOG FILE ABS. PATH >'

lisp.logging.cherrypy_access.level: 'ERROR'
lisp.logging.cherrypy_error.level: 'ERROR'
lisp.logging.cherrypy_app.level: 'DEBUG'
lisp.logging.cherrypy_console.level: 'ERROR'

lisp.verbose_errors_responses: False

"""

APIKEY_CONF_FOLDER = "conf.d"

APIKEY_CONF_EXAMPLE = """
# CONFIGURAZIONE DI BASE

[DEFAULT]
# apikey
key = <SET API KEY HERE>

# ip autorizzati
allow-ip = 127.0.0.1
allow-ip = 1.1.1.1
owner = pippo
group = pippo

actions-shell = /bin/bash
actions-timeout = 100

dir-perms = 0750
file-perms = 0640

##########################################
########## ALIASES DEPLOYMENT ############
##########################################

[ALIASES]

BASE = /home/pippo

##########################################
########## DIRECTORY DEPLOYMENT ##########
##########################################

[folder:/tmp]
# the files written in this directory will have the following owner/group and permissions
dir-perms = 0750

[folder:/home/pippo/Downloads]
# the files written in this directory will have the following owner/group and permissions
owner = redis
group = redis

file-perms = 0640

[folder:##BASE##/Downloads/folder1]
[folder:##BASE##/Downloads/fodler2/]

##########################################
########### AZIONI DEPLOYMENT ############
##########################################

[action:riavvia_nginx]
# commands executed on the server when the action is called
service nginx restart

[action:migra_database]
# commands executed on the server when the action is called
echo "Migrating database!"

"""




import os
import json
import logging
import cherrypy

from . import globals
from .functions import build_response, get_payload_data, log_api_call


class File(object):
    """
    Api server per il deployment
    """

    @staticmethod
    def __get_and_validate_perms(value):
        """
        Funzione di controllo dei permessi ottenuti via POST e ritorna i permessi utilizzabili dal deployer
        :param value:                   permessi indicati come stringa contenente un numero ottale
        :throw ValueError:              nel caso il permesso non sia un numero ottale
        :return:                        valore dei permessi come intero decimale
        """
        try:
            permissions = int(value, 8)
        except ValueError as e:
            # il numero passato non è convertibile
            raise ValueError('Permissions must be specified as an octal number (eg. 0o777, 0o0777, 0777, 777, etc.). ValueError: {}'.format(e))

        return permissions

    @cherrypy.expose(['upload'])
    def upload_file(self, *args, **kargs):
        """
        Vista per il caricamento di un file a chunk
        Sono ammesse solo chiamate POST
        :POST param apikey:     apikey che vuole effettuare l'azione
        :POST param file:       file da caricare
        :POST param dest_path:  destinazione del file (comprensiva del nome)
        :POST param perms:      permessi da impostare sul file (default = None)
        :POST param overwrite:  indica se eliminare il file nel caso esistesse (default = False)
        """

        verbose = cherrypy.config.get('lisp.verbose_errors_responses', False)

        if cherrypy.request.method != 'POST':
            # controllo se è una chiamata GET, in tal caso specifico che sono ammesse solo chiamate POST
            res = build_response(status=1, substatus=0, message="Only POST admitted")
            log_api_call(request=cherrypy.request, data='(omitted)', result=res)
            return res

        # controllo la presenza di tutti i campi necessari
        requested_fields = ['apikey', 'file', 'dest_path']
        if not all(k in kargs for k in requested_fields):
            res = build_response(status=1, substatus=4, message="Missing fields in JSON" if verbose else '', data=requested_fields if verbose else '')
            log_api_call(request=cherrypy.request, data=kargs, result=res)
            return res

        apikey = kargs['apikey']
        input_handle = kargs['file'].file
        dest_path = kargs['dest_path']
        overwrite_flag = kargs['overwrite'].lower().strip() == 'true' if 'overwrite' in kargs else False

        # provo ad ottenere i permessi dalla richiesta, mi arrivano come stringa contenente un ottale
        if 'perms' in kargs and kargs['perms'] is not None and len(kargs['perms']) > 0:
            try:
                # provo ad ottenere i permessi
                perms = self.__get_and_validate_perms(kargs['perms'])
            except ValueError as e:
                # nel caso non riesca ad ottenere i permessi ma questi ci siano, segnalo l'errore al client
                res = build_response(status=1, substatus=5, message="ValueError in request data", data="{}".format(e))
                log_api_call(request=cherrypy.request, data=kargs, result=res)
                return res
        else:
            # non sono stati specificati dei permessi, o sono vuoti
            perms = None

        # ottengo il file handle per poter scrivere il file sul server
        try:
            deployer, allowed_ips = globals.DEPLOYERS.get(apikey, (None, None))
            if deployer is None:
                res = build_response(status=2, substatus=1, message="Invalid Apikey")
                log_api_call(request=cherrypy.request, data=kargs, result=res)
                return res

            if allowed_ips is not None:
                if cherrypy.request.remote.ip not in allowed_ips:
                    res = build_response(status=2, substatus=6, message="Client ip not allowed")
                    log_api_call(request=cherrypy.request, data=kargs, result=res)
                    return res

            output_handle, output_file_path = deployer.get_file_uploader(filepath=dest_path, overwrite=overwrite_flag, permissions=perms)
        except PermissionError as e:
            res = build_response(status=2, substatus=2, message="Action not permitted", data='PermissionError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=kargs, result=res)
            return res
        except RuntimeError as e:
            res = build_response(status=2, substatus=3, message="A RuntimeError occurred", data='RuntimeError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=kargs, result=res)
            return res
        except FileExistsError as e:
            res = build_response(status=2, substatus=4, message="File already exists", data='FileExistsError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=kargs, result=res)
            return res

        # se sono arrivato qua sono sicuro che il file handle di output è valido

        # caricamento del file per chunk

        # inizializzo i valori per la massima dimensione del file e il chunk utilizzato
        max_size_mbytes = cherrypy.config.get('lisp.file_upload.max_size_mb', 100)
        max_size_bytes = max_size_mbytes * 1000 * 1000
        chunk_size = cherrypy.config.get('lisp.file_upload.chunk_size_mb', 5) * 1024 * 1024

        while True:
            # leggo un chunk alla volta
            data = input_handle.read(chunk_size)
            if not data:
                # non ci sono più dati da scrivere, ho finito
                break
            # scrivo il file
            output_handle.write(data)

            if output_handle.tell() > max_size_bytes:
                # il file ha sforato la dimensione massima

                # chiudo gli handle aperti dei file
                input_handle.close()
                output_handle.close()

                try:
                    # provo a rimuovere il file scritto parzialmente
                    os.remove(output_file_path)
                except (OSError, Exception) as e:
                    logging.error('Something went wrong deleting partial uploaded file "{}". Exception: {}'.format(output_file_path, e))

                # ritorno l'errore al chiamante
                res = build_response(status=2, substatus=5, message="File exceed maximum dimension", data="Max size: {} MB".format(max_size_mbytes))
                log_api_call(request=cherrypy.request, data=kargs, result=res)
                return res

        # arrivato qua il file è stato completamente caricato
        # chiudo gli handle
        input_handle.close()
        output_handle.close()

        res = build_response(status=0, substatus=0, message="File uploaded correctly")
        log_api_call(request=cherrypy.request, data=kargs, result=res)
        return res

    @cherrypy.expose(['mkdir'])
    @cherrypy.tools.clear_content_type_parsers()
    def create_directory(self):
        """
        creazione di una cartella
        Sono ammesse solo chiamate POST
        :JSON param apikey:     apikey che vuole effettuare l'azione
        :JSON param path:       path della cartella da creare
        :JSON param perms:      permessi della cartella da creare (default = None)
        :JSON param overwrite:  indica se sovrascrivere la cartella nel caso esistesse (default = False)
        """
        if cherrypy.request.method != 'POST':
            # controllo se è una chiamata GET, in tal caso specifico che sono ammesse solo chiamate POST
            res = build_response(status=1, substatus=0, message="Only POST admitted")
            log_api_call(request=cherrypy.request, data='(omitted)', result=res)
            return res

        try:
            # ottengo i dati della richiesta
            requested_fields = ['apikey', 'path']
            payload = get_payload_data(request=cherrypy.request, requested_fields=requested_fields)
        except RuntimeError as e:
            res = '{}'.format(e)
            log_api_call(request=cherrypy.request, data='(omitted)', result=res)
            return res

        apikey = payload['apikey']
        path = payload['path']
        overwrite = payload['overwrite'].lower().strip() == 'true' if 'overwrite' in payload else False

        # provo ad ottenere i permessi dalla richiesta, mi arrivano come stringa contenente un ottale
        if 'perms' in payload and payload['perms'] is not None and len(payload['perms']) > 0:
            try:
                # provo ad ottenere i permessi
                perms = self.__get_and_validate_perms(payload['perms'])
            except ValueError as e:
                # nel caso non riesca ad ottenere i permessi ma questi ci siano, segnalo l'errore al client
                res = build_response(status=1, substatus=5, message="ValueError in request data", data="{}".format(e))
                log_api_call(request=cherrypy.request, data=payload, result=res)
                return res
        else:
            # non sono stati specificati dei permessi, o sono vuoti
            perms = None

        try:
            deployer, allowed_ips = globals.DEPLOYERS.get(apikey, (None, None))
            if deployer is None:
                res = build_response(status=2, substatus=1, message="Invalid Apikey")
                log_api_call(request=cherrypy.request, data=payload, result=res)
                return res

            if allowed_ips is not None:
                if cherrypy.request.remote.ip not in allowed_ips:
                    res = build_response(status=2, substatus=6, message="Client ip not allowed")
                    log_api_call(request=cherrypy.request, data=payload, result=res)
                    return res

            deployer.create_folder(folderpath=path, overwrite=overwrite, permissions=perms)
        except PermissionError as e:
            res = build_response(status=2, substatus=2, message="Action not permitted", data='PermissionError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res
        except RuntimeError as e:
            res = build_response(status=2, substatus=3, message="A RuntimeError occurred", data='RuntimeError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res
        except FileExistsError as e:
            res = build_response(status=2, substatus=4, message="Folder already exists", data='FileExistsError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res

        res = build_response(status=0, substatus=0, message="Folder created")
        log_api_call(request=cherrypy.request, data=payload, result=res)
        return res

    @cherrypy.expose('delete_node')
    @cherrypy.tools.clear_content_type_parsers()
    def delete_filesystem_node(self):
        """
        eliminazione di un nodo del filesystem
        Sono ammesse solo chiamate POST
        :JSON param apikey:     apikey che vuole effettuare l'azione
        :JSON param path:       path del nodo da eliminare
        """
        if cherrypy.request.method != 'POST':
            # controllo se è una chiamata GET, in tal caso specifico che sono ammesse solo chiamate POST
            res = json.dumps({"status": "1", "message": "Only POST admitted", "data": None})
            log_api_call(request=cherrypy.request, data='(omitted)', result=res)
            return res

        try:
            # ottengo i dati della richiesta
            requested_fields = ['apikey', 'path']
            payload = get_payload_data(request=cherrypy.request, requested_fields=requested_fields)
        except RuntimeError as e:
            res = '{}'.format(e)
            log_api_call(request=cherrypy.request, data='(omitted)', result=res)
            return res

        apikey, path = payload['apikey'], payload['path']

        try:
            deployer, allowed_ips = globals.DEPLOYERS.get(apikey, (None, None))
            if deployer is None:
                res = build_response(status=2, substatus=1, message="Invalid Apikey")
                log_api_call(request=cherrypy.request, data=payload, result=res)
                return res

            if allowed_ips is not None:
                if cherrypy.request.remote.ip not in allowed_ips:
                    res = build_response(status=2, substatus=6, message="Client ip not allowed")
                    log_api_call(request=cherrypy.request, data=payload, result=res)
                    return res

            deployer.delete_filesystem_node(to_delete_path=path)
        except PermissionError as e:
            res = build_response(status=2, substatus=2, message="Action not permitted", data='PermissionError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res
        except RuntimeError as e:
            res = build_response(status=2, substatus=3, message="A RuntimeError occurred", data='RuntimeError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res
        except FileNotFoundError:
            res = build_response(status=0, substatus=1, message="Filesystem node not found")
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res

        res = build_response(status=0, substatus=0, message="Filesystem node deleted")
        log_api_call(request=cherrypy.request, data=payload, result=res)
        return res

    @cherrypy.expose('delete_by_regex')
    @cherrypy.tools.clear_content_type_parsers()
    def delete_by_regex(self):
        """
        eliminazione di un nodo del filesystem
        Sono ammesse solo chiamate POST
        :JSON param apikey:                apikey che vuole effettuare l'azione
        :JSON param path:                  path del nodo da eliminare
        :JSON param regex:                 regex con cui verificare se eliminare un elemento
        :JSON param inverted:              parametro che indica se invertire il matche effettuato dalla regex
        :JSON param node_type:             indica se considerare solo file, solo cartelle o entrambi nell'eliminazione
        :JSON param include_subdirs:       indica se considerare solo la directory corrente o anche le sottodirectory per le eliminazioni

        """
        if cherrypy.request.method != 'POST':
            # controllo se è una chiamata GET, in tal caso specifico che sono ammesse solo chiamate POST
            res = json.dumps({"status": "1", "message": "Only POST admitted", "data": None})
            log_api_call(request=cherrypy.request, data='(omitted)', result=res)
            return res

        try:
            # ottengo i dati della richiesta
            requested_fields = ['apikey', 'path', 'regex', 'node_type']
            payload = get_payload_data(request=cherrypy.request, requested_fields=requested_fields)
        except RuntimeError as e:
            res = '{}'.format(e)
            log_api_call(request=cherrypy.request, data='(omitted)', result=res)
            return res

        apikey, path = payload['apikey'], payload['path']
        regex, node_type = payload['regex'], payload['node_type'].lower().strip(),
        inverted = payload['inverted'] if 'inverted' in payload else False
        include_subdirs = payload['include_subdirs'] if 'include_subdirs' in payload else False

        node_type_allowed_values = {'file', 'folder', 'both'}
        if node_type not in node_type_allowed_values:
            res = build_response(status=2, substatus=5, message="ValueError in JSON",  data='The value of node_type parameter is invalid. Allowed values: {}'.format(node_type_allowed_values))
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res

        try:
            deployer, allowed_ips = globals.DEPLOYERS.get(apikey, (None, None))
            if deployer is None:
                res = build_response(status=2, substatus=1, message="Invalid Apikey")
                log_api_call(request=cherrypy.request, data=payload, result=res)
                return res

            if allowed_ips is not None:
                if cherrypy.request.remote.ip not in allowed_ips:
                    res = build_response(status=2, substatus=6, message="Client ip not allowed")
                    log_api_call(request=cherrypy.request, data=payload, result=res)
                    return res

            errors, too_many_errors = deployer.delete_by_regex(base_path=path, regex=regex, inverted=inverted, node_type=node_type, include_subdirs=include_subdirs)
        except PermissionError as e:
            res = build_response(status=2, substatus=2, message="Action not permitted", data='PermissionError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res
        except RuntimeError as e:
            res = build_response(status=2, substatus=3, message="A RuntimeError occurred", data='RuntimeError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res
        except FileNotFoundError:
            res = build_response(status=0, substatus=1, message="Filesystem node not found")
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res

        res = build_response(status=0, substatus=0 if len(errors) == 0 else 3, message="Nodes deleted successfully" if len(errors) == 0 else "Some errors occurred", data={'errors': errors, 'too_many_errors': too_many_errors})
        log_api_call(request=cherrypy.request, data=payload, result=res)
        return res


import sys
import cherrypy
import argparse
from .functions import file_check, folder_check, create_config_dir, generate_apikey
from .kaomi_server import start_server


# parser per gli argomenti via linea di comando
arg_parser = argparse.ArgumentParser(prog='python -m kaomi_server', description="Kaomi server (TM)")


subparsers = arg_parser.add_subparsers(help='help for subcommand', dest='subcommand', required=True)

# create the parser for the "start" command
parser_start = subparsers.add_parser('start', help='Starts the server')
parser_start.add_argument('--config', type=file_check, dest='server_config_path', metavar="config_path", required=True, help="Path of the configuration file for the server")
parser_start.add_argument('--apikey', type=folder_check, dest='apikey_configs_folder', metavar="apikey_configs_path", required=True, help="Path of the folder containing apikey's configs")

# create the parser for the "configure" command
parser_configure = subparsers.add_parser('configure', help='Configure a directory to store configuration files')
parser_configure.add_argument('--folder', type=folder_check, dest='configs_folder', metavar="configs_folder", required=True, help='Path where to create the example config for server and api keys')

# create the parser for the "apikey" command
parser_apikey = subparsers.add_parser('gen-api-key', help='Prints a secure apikey for the configuration')

# parso gli argomenti passati alla chiamata
args = arg_parser.parse_args(sys.argv[1:])

# svuoto il timestamp dal logging di cherrypy perché non venga stampato due volte
cherrypy._cplogging.LogManager.time = lambda self: ""

if args.subcommand == 'configure':
    try:
        create_config_dir(args.configs_folder)
    except RuntimeError as e:
        print("{}".format(e))

elif args.subcommand == 'start':
    start_server(config_file=args.server_config_path, apikey_folder=args.apikey_configs_folder)
elif args.subcommand == 'gen-api-key':
    print(generate_apikey())
else:
    print("The subcommand is invalid. This should not happen. Bug.")



import gzip
import os
import shutil

from logging import handlers

class CompressingRotatingFileHandler(handlers.RotatingFileHandler):
    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)

    def doRollover(self):
        if self.stream:
            self.stream.close()
            self.stream = None

        if self.backupCount > 0:
            for i in range(self.backupCount - 1, 0, -1):
                sfn = self.rotation_filename("%s.%d.gz" % (self.baseFilename, i))
                dfn = self.rotation_filename("%s.%d.gz" % (self.baseFilename, i + 1))
                if os.path.exists(sfn):
                    if os.path.exists(dfn):
                        os.remove(dfn)
                    os.rename(sfn, dfn)
            dfn = self.rotation_filename(self.baseFilename + ".1")
            if os.path.exists(dfn):
                os.remove(dfn)
            self.rotate(self.baseFilename, dfn)
        if not self.delay:
            self.stream = self._open()


        log_dir = os.path.dirname(self.baseFilename)

        to_compress = [
            os.path.join(log_dir, f)
            for f in os.listdir(log_dir)
            if f.startswith(os.path.basename(os.path.splitext(self.baseFilename)[0])) and not f.endswith((".gz", ".log"))
        ]

        for f in to_compress:
            if os.path.exists(f):
                with open(f, "rb") as _old, gzip.open(f + ".gz", "wb") as _new:
                    shutil.copyfileobj(_old, _new)
                os.remove(f)


import logging
import sys

import cherrypy
import logging.config


# registro i tool prima di importare le viste che ne fanno uso
from .globals import SERVER_VERSION


@cherrypy.tools.register('before_request_body')
def clear_content_type_parsers():
    request = cherrypy.serving.request
    request.body.processors.clear()


from .views_files import File
from .views_alive import Alive
from .views_actions import Action
from . import globals
from .functions import load_apikey_configs, err_404_error_handler, default_error_handler


def setup_logging():
    """
    Procdeura di inizializzazione e setup del logging in base alle configurazioni
    """
    # imposto i nomi dei file di logging come sono scritti nella configurazione di cherrypy
    globals.LOG_CONF['handlers']['cherrypy_access']['filename'] = cherrypy.config.get('lisp.logging.cherrypy_access.file', 'server_access.log')
    globals.LOG_CONF['handlers']['cherrypy_error']['filename'] = cherrypy.config.get('lisp.logging.cherrypy_error.file', 'server_error.log')
    globals.LOG_CONF['handlers']['cherrypy_app']['filename'] = cherrypy.config.get('lisp.logging.cherrypy_app.file', 'app.log')

    # imposto i livelli di logging in base alla configurazione
    globals.LOG_CONF['handlers']['cherrypy_access']['level'] = cherrypy.config.get('lisp.logging.cherrypy_access.level', 'INFO')
    globals.LOG_CONF['handlers']['cherrypy_error']['level'] = cherrypy.config.get('lisp.logging.cherrypy_error.level', 'INFO')
    globals.LOG_CONF['handlers']['cherrypy_error']['level'] = cherrypy.config.get('lisp.logging.cherrypy_app.level', 'DEBUG')
    globals.LOG_CONF['handlers']['cherrypy_console']['level'] = cherrypy.config.get('lisp.logging.cherrypy_console.level', 'INFO')
    globals.LOG_CONF['handlers']['default']['level'] = cherrypy.config.get('lisp.logging.default.level', 'INFO')

    # setup del logging utilizzando la configurazione su file
    logging.config.dictConfig(globals.LOG_CONF)




def start_server(config_file, apikey_folder):
    """
    Procedura per l'avvio del server
    :param config_file:         path del file di configurazione del server
    :param apikey_folder:       path della cartella in cui sono presenti le configurazioni delle apikey
    """
    # ottengo la configurazione del server dal file passato
    cherrypy.config.update(config_file)
    # imposto l'handler di errori del server
    cherrypy.config.update({
        'error_page.404': err_404_error_handler,
        'error_page.default': default_error_handler,
        'response.headers.server': '',
    })

    # faccio il setup del logging
    setup_logging()

    try:
        globals.DEPLOYERS = load_apikey_configs(apikey_folder=apikey_folder)
    except RuntimeError as e:
        print('Something went wrong loading configs. {}'.format(e))
        sys.exit(1)

    # mounting classes
    cherrypy.tree.mount(Alive(), '/', config_file)
    cherrypy.tree.mount(File(), '/api/v1/files/', config_file)
    cherrypy.tree.mount(Action(), '/api/v1/actions/', config_file)

    signal_handler = cherrypy.process.plugins.SignalHandler(bus=cherrypy.engine)
    signal_handler.subscribe()



    # starting server
    cherrypy.engine.start()

    l = logging.getLogger('app')
    l.info('Started kaomi-server {}'.format(SERVER_VERSION))


    cherrypy.engine.block()



import copy
import logging
import os
import json
import argparse
import secrets
import string

import cherrypy
from .example_files import APIKEY_CONF_FOLDER, SERVER_CONF_EXAMPLE, APIKEY_CONF_EXAMPLE
from .deployer import Deployer
from .config_file import ConfigFile

def folder_check(path):
    """ Funzione di controllo delle cartelle passate via linea di comando """
    if not os.path.isabs(path):
        # path non assoluto
        raise argparse.ArgumentTypeError('The path must be absolute')
    if not os.path.isdir(path):
        # file non trovato
        raise argparse.ArgumentTypeError('Cannot find specified folder {}'.format(path))
    return path


def file_check(path):
    """ Funzione di controllo dei file passati via linea di comando """
    if not os.path.isabs(path):
        # path non assoluto
        raise argparse.ArgumentTypeError('The path must be absolute')
    if not os.path.isfile(path):
        # file non trovato
        raise argparse.ArgumentTypeError('Cannot find specified file {}'.format(path))
    return path


def get_payload_data(request, requested_fields):
    """
    Funzione che data la richiesta restituisce il dizionario contenente il payload parsato
    :param request:             richiesta
    :param requested_fields:    campi che devono essere presenti
    :throw RuntimeError:        nel caso si verificasse qualche problema con la richiesta
                                nel caso venga sollevata un'eccezione, conterrà una stringa contenente il dizionario da restituire al chiamante
    :return:                    dizionario contenente i dati
    """

    verbose = cherrypy.config.get('lisp.verbose_errors_responses', False)

    try:
        # ottengo la dimensione del payload
        content_length = int(request.headers['Content-Length'])

        # ottengo la dimensione massima ammissibile (definita ad hoc) dalla configurazione
        max_request_size = cherrypy.config.get('lisp.request.max_size_mb', 2) * 1000**2

        # controllo sulla dimensione massima ammissibile
        if content_length > max_request_size:
            # la dimensione della richiesta supera la dimensione massima consentita
            raise RuntimeError(build_response(status=1, substatus=1, message="Payload content-length greater than maximum allowed", data="Max size: {} MB".format(max_request_size) if verbose else ''))

        # per ovviare al fatto che la grandezza del payload potrebbe non coincidere con quella dichiarata nel content-length,
        # andiamo a leggere solo content_length dati
        payload = json.loads(request.body.read(content_length))
    except json.JSONDecodeError:
        # errore nella decodifica del json
        raise RuntimeError(build_response(status=1, substatus=2, message="Json content cannot be parsed" if verbose else ''))
    except ValueError:
        # errore nel casting per il content_length
        raise RuntimeError(build_response(status=1, substatus=3, message="Content-length not specified or not valid integer" if verbose else ''))

    # controllo che tutti i campi richiesti siano presenti nel payload
    if not all(k in payload for k in requested_fields):
        raise RuntimeError(build_response(status=1, substatus=4, message="Missing fields in JSON" if verbose else '', data=requested_fields if verbose else None))

    return payload


def build_response(status, substatus, message="", data=None):
    """
    Funzione per la creazione della risposta del server
    :param status:      stato della richiesta (int)
    :param substatus:   sotto-stato della richiesta (int)
    :param message:     messaggio descrittivo del sottostato
    :param data:        eventuali dati di output
    """
    return json.dumps({
        "status": status,
        "substatus": substatus,
        "message": message,
        "data": data
    })


def create_config_dir(path):
    """
    Funzione per la creazione dei file di configurazione di esempio del server e delle apikey
    :param path:            cartella nella quale vengono creati gli esempi
    :throw RuntimeError:    nel caso si verifichi un errore
    """
    if not os.path.isdir(path):
        # la cartella non si trova
        raise RuntimeError('Cannot find specified directory "{}".'.format(path))

    # creo la directory per le configurazioni dell'apikey
    apikey_conf_folder = os.path.join(path, APIKEY_CONF_FOLDER)
    try:
        # la cartella non viene sovrascritta, nel caso esista si deve eliminarla
        os.mkdir(apikey_conf_folder)
    except (OSError, FileExistsError) as e:
        raise RuntimeError('Something went wrong creating folder {}. Error: {}'.format(apikey_conf_folder, e))

    # ottengo i path dei file di esempio
    server_conf_path = os.path.join(path, 'server.conf.example')
    apikey_conf_path = os.path.join(apikey_conf_folder, 'dashboard.ini.example')

    # scrivo i file di esempio
    try:
        # i file nel caso dovessero esistere vengono sovrascritti

        with open(server_conf_path, 'w') as fh:
            # file conf server
            fh.write(SERVER_CONF_EXAMPLE)
        with open(apikey_conf_path, 'w') as fh:
            # file conf apikey
            fh.write(APIKEY_CONF_EXAMPLE)
    except OSError as e:
        # nel caso qualcosa vada storto con la scrittura dei file segnalo il problema al chiamante
        raise RuntimeError('Something went wrong trying to create example files. Error: {}'.format(e))


def default_error_handler(status, message, traceback, version):
    """
    Funzione per la mascheratura degli errori del server (500, etc)
    """
    # gli errori vengono loggati in automatico da cherrypy

    logging.info("Default error handler called: status={}, message='{}', traceback:{}".format(status, message, traceback))

    # imposto lo stato della risposta
    cherrypy.response.status = 500

    # imposto il contenuto della risposta
    return json.dumps({'status': 2, 'substatus': 0, 'message': 'Oops, something went wrong', 'data': None})

    # volendo si possono fare altre azioni come inviare un email

def err_404_error_handler(status, message, traceback, version):
    """
    Funzione per gestire gli errori 404
    """
    # gli errori vengono loggati in automatico da cherrypy

    logging.info("Default error handler called: {} {} {}".format(status, message, traceback))

    # imposto lo stato della risposta
    cherrypy.response.status = 404

    # imposto il contenuto della risposta
    return json.dumps({'status': 2, 'substatus': 0, 'message': 'The requested path does not exist.'})


def load_apikey_configs(apikey_folder):
    """
    Procedura per l'ottenimento del dizionario
    :param apikey_folder:   cartella in cui sono presenti le configs
    :throw RuntimeError:    cartella non esiste, errore nel parse delle config, etc.
    :return:                dizionario con i deployers { 'apikey' : <Deployer> }
    """
    # dizionario che verrà restituito come output
    deployers_dict = {}

    l = logging.getLogger('app')

    loaded_configs = []

    try:
        # lettura delle configurazioni dalla cartella delle config
        for config_file in os.listdir(apikey_folder):
            # controllo che la configurazione abbia il formato corretto
            if os.path.splitext(config_file)[-1].lower() == '.ini':
                # inizializzo un oggetto configurazione
                this_config = ConfigFile(os.path.join(apikey_folder, config_file))
                this_config_dict = this_config.to_dict()

                this_apikey = this_config_dict['default']['key']

                # creo il deployer per questa configurazione
                deployers_dict[this_apikey] = Deployer(configuration=this_config_dict), this_config_dict['default'].get('allow-ip', None)

                # il caricamento ha avuto successo
                loaded_configs.append(os.path.basename(config_file))
            else:
                # altrimenti ignoro il file
                continue
        # finito il caricamento di tutte le config, restituisco il dizionario con i deployer
        l.info('Api configurations loaded successfully: {}'.format(', '.join(loaded_configs)))

        return deployers_dict
    except (ValueError, FileNotFoundError, RuntimeError) as e:
        # try/except esterno al for in modo da bloccare il funzionamento se esiste anche un'unica config errata
        l.error('An error occurred loading api key configs. Error {}: {}'.format(type(e), e))
        raise RuntimeError('An error occurred loading api key configs. Error {}: {}'.format(type(e), e))


def log_api_call(request, data, result):
    l = logging.getLogger('app')

    if isinstance(data, dict):
        # evito di modifcare il dizionario del chiamante
        data = data.copy()
        # se è presente un api key la tolgo
        if 'apikey' in data:
            data['apikey'] = '(omitted)'

        data = json.dumps(data, default=str)

    if len(data) > 300:
        data = data[:300] + '...(truncated)'

    if isinstance(result, dict):
        result = json.dumps(result, default=str)

    if len(result) > 300:
        result = result[:300] + '...(truncated)'

    l.debug('{} {} {} {}'.format(request.remote.ip, cherrypy.url(qs=cherrypy.request.query_string, relative="server"), data, result))


def generate_apikey():
    alphabet = string.ascii_letters + string.digits
    generated_key = ''.join(secrets.choice(alphabet) for i in range(48))
    return generated_key




from . import __version__

# dizionario che conterrà i deployers
DEPLOYERS = {}


# Dizionario contenente la configurazione del logging
# Nota: i file di output vengono sovrascritti all'accensione del server da quelli specificati nella configurazione kaomi

LOG_CONF = {
    'version': 1,

    'formatters': {
        'void': {
            'format': ''
        },
        'standard': {
            'format': '[%(asctime)s][%(levelname)s][%(name)s] %(message)s'
        },
        'console': {
            'format': '[%(asctime)s][%(levelname)s] %(message)s'
        }
    },
    
    'handlers': {
        'default': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'standard',
            'stream': 'ext://sys.stdout'
        },
        'cherrypy_console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'console',
            'stream': 'ext://sys.stdout'
        },
        'cherrypy_access': {
            'level': 'INFO',
            'class': 'kaomi_server.log.CompressingRotatingFileHandler',
            'formatter': 'standard',
            'filename': 'access.log',
            'maxBytes': 52428800,
            'backupCount': 100,
            'encoding': 'utf8'
        },
        'cherrypy_error': {
            'level': 'INFO',
            'class': 'kaomi_server.log.CompressingRotatingFileHandler',
            'formatter': 'standard',
            'filename': 'error.log',
            'maxBytes': 52428800,
            'backupCount': 100,
            'encoding': 'utf8'
        },
        'cherrypy_app': {
            'level': 'DEBUG',
            'class': 'kaomi_server.log.CompressingRotatingFileHandler',
            'formatter': 'standard',
            'filename': 'app.log',
            'maxBytes': 52428800,
            'backupCount': 100,
            'encoding': 'utf8'
        },
    },

    'loggers': {
        '': {
            'handlers': ['cherrypy_app'],
            'level': 'DEBUG'
        },
        'cherrypy.app': {
            'handlers': ['cherrypy_app'],
            'level': 'DEBUG'
        },
        'cherrypy.access': {
            'handlers': ['cherrypy_access'],
            'level': 'DEBUG',
            'propagate': False
        },
        'cherrypy.error': {
            'handlers': ['cherrypy_console', 'cherrypy_error'],
            'level': 'DEBUG',
            'propagate': False
        },
    }
}


# la versione del server viene presa dall'init
SERVER_VERSION = __version__


import cherrypy
from .functions import build_response, log_api_call


class Alive(object):
    """
    Classe contenitore di viste che danno informazioni di stato ai client
    """

    @cherrypy.expose()
    @cherrypy.tools.clear_content_type_parsers()
    def index(self):
        """
        Utilizzato per indicare se il kaomi deployer è up
        """
        res = build_response(status=0, substatus=0, message="https://www.youtube.com/watch?v=oHg5SJYRHA0")
        log_api_call(request=cherrypy.request, data='(omitted)', result=res)
        return res



import json
import logging

import cherrypy
from .deployer import ActionNotFoundError
from .functions import get_payload_data, build_response, log_api_call
from . import globals


class Action(object):
    """
    Classe contenitore delle chiamate relative alle azioni
    """

    @cherrypy.expose(['execute'])
    @cherrypy.tools.clear_content_type_parsers()
    def execute_action(self):
        """
        esecuzione di un'azione
        Sono ammesse solo chiamate POST
        :POST param apikey:     apikey che vuole effettuare l'azione
        :POST param action:     nome dell'azione che si vuole eseguire
        """
        if cherrypy.request.method != 'POST':
            # controllo se è una chiamata GET, in tal caso specifico che sono ammesse solo chiamate POST
            res = json.dumps({"status": "1", "message": "Only POST admitted", "data": None})
            log_api_call(request=cherrypy.request, data='(omitted)', result=res)
            return res

        try:
            # ottengo i dati della richiesta
            requested_fields = ['apikey', 'action']
            payload = get_payload_data(request=cherrypy.request, requested_fields=requested_fields)
        except RuntimeError as e:
            res = '{}'.format(e)
            log_api_call(request=cherrypy.request, data='(omitted)', result=res)
            return res

        apikey, action = payload['apikey'], payload['action']

        try:
            deployer, allowed_ips = globals.DEPLOYERS.get(apikey, (None, None))
            if deployer is None:

                res = build_response(status=2, substatus=1, message="Invalid Apikey")
                log_api_call(request=cherrypy.request, data=payload, result=res)
                return res

            if allowed_ips is not None:
                if cherrypy.request.remote.ip not in allowed_ips:
                    res = build_response(status=2, substatus=6, message="Client ip not allowed")
                    log_api_call(request=cherrypy.request, data=payload, result=res)
                    return res

            action_result = deployer.execute_action(action)
        except ActionNotFoundError as e:
            res = build_response(status=2, substatus=2, message="Action not permitted", data='ActionNotFoundError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res

        except RuntimeError as e:
            res = build_response(status=2, substatus=3, message="A RuntimeError occurred", data='RuntimeError: {}'.format(e))
            log_api_call(request=cherrypy.request, data=payload, result=res)
            return res

        res = build_response(status=0, substatus=0, message="Action executed", data=action_result)
        log_api_call(request=cherrypy.request, data=payload, result=res)
        return res

